/*
    WN/SSL
    File: wnssl.c
    Version 2.0-1

    Copyright (C) 1996  <by Elias Doumas>
    Copyright (C) 1998-2000 <by Matthias Cramer>
    
    In appreciation of the work by Tim Hudson which has been the foundation
    for this developement.

    Modified for WN 2.0 and OpenSSL by Matthias Cramer <cramer@freestone.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

CHANGES:

 V2.0-1
   21-Sep-1999 Mathias Cramer <cramer@freestone.net> 
     - Added call to SSL_CTX_set_tmp_rsa_callback() to support
       40 bit encryption. (Thanks to Andrew Morrow)

*/

#include "wnssl.h"

char ssl_file_path[MIDLEN],
     ssl_buf[MIDLEN],
     env_cipher[MIDLEN],
     env_subject[MIDLEN],
     env_issuer[MIDLEN],
     env_serial[MIDLEN];

int  err;


#if OPENSSL_VERSION_NUMBER >= 0x00904000L
# define myPEM_read_RSAPrivateKey(f,k,x)   PEM_read_RSAPrivateKey(f,k,x,NULL)
# define myPEM_read_X509(f,c,x)            PEM_read_X509(f,c,x,NULL)
#else
# define myPEM_read_RSAPrivateKey(f,k,x)   PEM_read_RSAPrivateKey(f,k,x)
# define myPEM_read_X509(f,c,x)            PEM_read_X509(f,c,x)
#endif


#ifdef USE_WN_READ
int
WN_read(a,b,c)
int     a;
char    *b;
int     c;
{
       int	ret_val;
       if (ssl_active_flag && (is_ssl_fd(ssl_con,(fileno(stdin)))))
          ret_val=SSL_read(ssl_con,b,c);
       else
          ret_val=read(a,b,c);
       return ret_val;
}
#endif

#ifdef USE_WN_WRITE
int
WN_write(a,b,c)
int     a;
char    *b;
int     c;
{
        int     ret_val;

        if (ssl_active_flag && (is_ssl_fd(ssl_con,a)))
                ret_val=SSL_write(ssl_con,b,c);
        else
                ret_val=write(a,b,c);
        return ret_val;
}
#endif


#if SSLEAY_VERSION_NUMBER >= 0x0800 && SSLEAY_VERSION_NUMBER < 0x0920
/* FIXME: This ought to generate a new key from time to time - why make the
Feds' life easy? */
static RSA *TmpRSACallback(SSL *pSSL,int nExport)
    {
    static RSA *pRSA=NULL;

    if (pRSA == NULL)
# if SSLEAY_VERSION_NUMBER >= 0x0900
        pRSA=RSA_generate_key(512,RSA_F4,NULL,NULL);
# else
        pRSA=RSA_generate_key(512,RSA_F4,NULL);
# endif
    return pRSA;
    } 
#endif

#if SSLEAY_VERSION_NUMBER >= 0x0920
/* FIXME: This ought to generate a new key from time to time - why make the
Feds' life easy? */
static RSA *TmpRSACallback(SSL *pSSL,int nExport,int nKeyLength)
    {
    static RSA *pRSA512=NULL; 
    static RSA *pRSA1024=NULL;

    assert(nExport);  /* Otherwise it was a sign-only key and we should set our keylength */
    assert(nKeyLength == 512 || nKeyLength == 1024);                           
    if (pRSA512 == NULL && nKeyLength == 512)
        pRSA512=RSA_generate_key(512,RSA_F4,NULL,NULL);
    if (pRSA1024 == NULL && nKeyLength == 1024)
        pRSA1024=RSA_generate_key(1024,RSA_F4,NULL,NULL);
    
    return nKeyLength == 512 ? pRSA512 : pRSA1024;
    } 
#endif



void wnssl_setup()
{
    FILE *errfp;
    X509 *cert;
    int  serial;

    if (ssl_only_flag)  {

   /* do the SSL stuff now ... before we play with pty's */

      ssl_con=(SSL *)SSL_new(ssl_ctx);
      SSL_set_fd(ssl_con,fileno(stdout));

      if (SSL_use_RSAPrivateKey(ssl_con,ssl_private_key)==0)  {
      	logerr("Error reading Private Key","");
      	exit(1);
      }

      if (SSL_use_certificate(ssl_con,ssl_public_cert)==0) {
        logerr("Error reading certificate","");
        exit(1);
      }

      SSL_set_verify(ssl_con,ssl_verify_flag,NULL);

      if (SSL_accept(ssl_con)<=0)  {
/*          logerr("Accept failed",""); */
/*          SSL_free(ssl_con); */
	while ((err = ERR_get_error()))
	  {
	    ERR_error_string(err, &ssl_buf);
	    logerr ("Access failed", ssl_buf);
	  }

         sleep(1);
         exit(1);
      } else  {
          ssl_active_flag=1;
      }

   }

#ifdef WNSSL_ENVIRON
    sprintf(env_cipher, "CC_CIPHER=%.256s", SSL_get_cipher(ssl_con));
    putenv (env_cipher);
    strcpy (ssl_buf, "");
    if ( (cert = SSL_get_peer_certificate( ssl_con))) {

      X509_NAME_oneline( X509_get_subject_name(cert), ssl_buf, 220);
      sprintf(env_subject, "CC_SUBJECT=%.256s", ssl_buf);
      putenv(env_subject);

      X509_NAME_oneline( X509_get_issuer_name(cert), ssl_buf, 220);
      sprintf(env_issuer, "CC_ISSUER=%.256s", ssl_buf);
      putenv(env_issuer);

      ASN1_INTEGER_set( X509_get_serialNumber(cert), serial);
      sprintf(env_serial, "CC_SERIAL=%.256s", ssl_buf);
      putenv(env_serial);

      X509_free (cert);
    }

#ifdef WNSSL_DEBUG
    errfp = fopen("/tmp/sslerror", "a+");
    fprintf( errfp, "%s\n", env_subject );
    fprintf( errfp, "%s\n", env_issuer );
    fprintf( errfp, "%s\n", env_serial );
    fclose( errfp );
#endif
#endif
      
/* ssl_fprintf calls require that this be null to test
 * for being an ssl stream
 */
   if (!ssl_only_flag || !ssl_active_flag)
     ssl_con=NULL;

/* Patch to reset the value for our new connection */
   this_conp->scheme = "https";
}

void wnssl_opts(optarg)
char *optarg;
{
           if (strcmp(optarg, "debug") == 0 )  {
              ssl_debug_flag=1;
           }
           if (strcmp(optarg, "ssl") == 0 )  {
              ssl_only_flag=1;
           }
           if (strcmp(optarg, "standalone") == 0 )  {
              standalone_debug=1;
           }
           if (strncmp(optarg, "verify=", strlen("verify=")) == 0 )  {
               ssl_verify_flag=atoi(optarg+strlen("verify="));
           }
           if (strncmp(optarg, "cert=", strlen("cert=")) == 0 )  {
               my_ssl_cert_file=optarg+strlen("cert=");
           }
           if (strncmp(optarg, "key=", strlen("key=")) == 0 )  {
               my_ssl_key_file=optarg+strlen("key=");
           }

}

void wnssl_init(port, errlogfp)
int *port;
FILE *errlogfp;
{
   if (ssl_only_flag)
     {
        FILE *fp;
        char *filename;

	SSLeay_add_ssl_algorithms();

        /* if we are not running in debug then any error
         * stuff from SSL debug *must* not go down
         * the socket (which 0,1,2 are all pointing to by
         * default)
         */

        /* init things so we will get meaningful error messages
         * rather than numbers
         */
        SSL_load_error_strings();

        ssl_ctx=(SSL_CTX *)SSL_CTX_new(SSLv23_server_method());

        if (!SSL_CTX_set_default_verify_paths(ssl_ctx))  {
          fprintf(stderr,"wnssl: cannot set default path via X509_set_default_verify_paths\n");
          fflush(stderr);
          exit(1);
        }

#if 0
        sprintf(ssl_file_path,"%s/%s",X509_get_default_cert_dir(),
                "wnssl.cert");
#endif

        sprintf(ssl_file_path,"%s/%s",X509_get_default_cert_dir(),
                "wnssl.pem");

        filename=my_ssl_cert_file==NULL?ssl_file_path:my_ssl_cert_file;
        fp=fopen(filename,"r");
        if (fp==NULL)  {
           fprintf(stderr,"wnssl: cannot open public cert file \"%s\"\n",filename);
           fflush(stderr);
           exit(1);
        }

        ssl_public_cert=X509_new();

        if (myPEM_read_X509(fp,&ssl_public_cert,NULL)==0)  {
           fprintf(stderr,"wnssl: error reading public cert - %s\n",
           ERR_error_string(err, ERR_get_error()));
           fflush(stderr);
           exit(1);
        }
        fclose(fp);

#if 0
        sprintf(ssl_file_path,"%s/private/%s",X509_get_default_cert_area(),
                "wnssl.key");
#endif

        sprintf(ssl_file_path,"%s/%s",X509_get_default_cert_dir(),
                "wnssl.pem");

        filename=my_ssl_key_file==NULL?ssl_file_path:my_ssl_key_file;
        fp=fopen(filename,"r");
        if (fp==NULL)  {
           fprintf(stderr,"wnssl: cannot open private key file \"%s\"\n",filename);
           fflush(stderr);
           exit(1);
        }

        ssl_private_key=RSA_new();
        if (myPEM_read_RSAPrivateKey(fp,&ssl_private_key,NULL)==0)  {
           fprintf(stderr,"wnssl: error reading private key - %s\n",
                   ERR_error_string(err, ERR_get_error()));
           fflush(stderr);
           exit(1);
        }
        fclose(fp);
        
        SSL_CTX_set_tmp_rsa_callback(ssl_ctx,TmpRSACallback);

     }
    /* override things ... we are sharing config files for
     * the moment
     */
    if (ssl_only_flag)
       *port = DEFAULT_SSL_PORT;
}
